#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#define DEBUG
#define DEBUG_ENTER
#define DEBUG_EXIT

typedef unsigned char byte;

struct block {
  size_t block_size;
  size_t used_size;
  struct block *next;
  int used;
};

struct block *first = NULL;

struct block *new_block(size_t size) {
  struct block *base;
#ifdef DEBUG
  fprintf(stderr,"**** NEW BLOCK ****\n");
#endif
  base = (struct block *)sbrk((int) (size + sizeof(struct block)));
  if (base == (struct block *) -1) {
    perror ("sbrk");
    return NULL;
  }
  base->next = NULL;
  base->used = 0;
  base->block_size = size;
  return base;
}


void *mmalloc (size_t size) {
  struct block *pblock;
  struct block *prev_pblock;

  if ( NULL == first ) {
    if ( NULL == (first = new_block(size)) ) {
      return NULL;
    }
    first->used = 1;
    first->used_size = size;
    return (void *) (first+1);
  }

  pblock = first;
  prev_pblock = first;
  while (NULL != pblock) {
    if ( !pblock->used && pblock->block_size >= size ) {
#ifdef DEBUG
      fprintf(stderr,"**** OLD BLOCK ****\n");
#endif
      pblock->used = 1;
      pblock->used_size = size;
      return (void *) (pblock+1);
    }
    prev_pblock = pblock;
    pblock = pblock->next;
  }

  /* Il n'y a pas de block libre */
  if ( NULL == (pblock = new_block(size)) ) {
    return NULL;
  }
  prev_pblock->next = pblock;
  pblock->used = 1;
  pblock->used_size = size;
  return (void *) (pblock+1);
}

void mfree(void *p) {
  /* TODO */
}

void * mrealloc(void *p, size_t size) {
  /* TODO */
  return NULL;
}

void * mcalloc(size_t count, size_t size) {
  /* TODO */
  return NULL;
}


/* Redefinition des fonctions malloc, free, realloc et calloc */

void *malloc (size_t size) {
  void *p;
#ifdef DEBUG_ENTER
  fprintf(stderr,"-> enter malloc\n");
#endif
  p = mmalloc(size);
#ifdef DEBUG_EXIT
  fprintf(stderr,"-> exit malloc\n");
#endif
  return p;
}

void free(void *p) {
#ifdef DEBUG_ENTER
  fprintf(stderr,"-> enter free\n");
#endif
  mfree(p);
#ifdef DEBUG_EXIT
  fprintf(stderr,"-> exit free\n");
#endif
}

void * realloc(void *ptr, size_t size) {
  void *p;
#ifdef DEBUG_ENTER
  fprintf(stderr,"-> enter realloc\n");
#endif
  p = mrealloc(ptr, size);
#ifdef DEBUG_EXIT
  fprintf(stderr,"-> exit realloc\n");
#endif
  return p;
}

void * calloc(size_t count, size_t size) {
  void *p;
#ifdef DEBUG_ENTER
  fprintf(stderr,"-> enter calloc\n");
#endif
  p = mcalloc(count, size);
#ifdef DEBUG_EXIT
  fprintf(stderr,"-> exit calloc\n");
#endif
  return p;
}
